/* StatServ routines

   OperStats - Copyright (c) 2000-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"

TLD *tldlist = NULL;
Zone *zonelist = NULL;
Boolean_T tldlookup = FALSE, zonelookup = FALSE;

static void do_help (const char *source);
static void do_stats (const char *source);
static void do_map (const char *source);
static void do_servers (const char *source);
static void do_tld (const char *source);
static void do_tldmap (const char *source);
static void do_tldstats (const char *source);
static void do_zone (const char *source);
static void do_reset (const char *source);
static void do_flush (const char *source);
static void do_sendpings (const char *source);
static void sh_search (const char *source);
static void sh_auth (const char *source);
static void sh_deauth (const char *source);
#ifdef RAWINJECT
static void sh_raw (const char *source);
static void sh_inject (const char *source);
#endif
static void sh_debug (const char *source);
static void sh_shutdown (const char *source);
static void sh_restart (const char *source);
static void sh_settings (const char *source);
static void sh_update (const char *source);
static void sh_rehash (const char *source);

static int showleafs (const char *source, char *server, char *prevpre, char *arrowchar);
static int numleafs (char *server);

/* Main StatServ routine. */
void statserv (const char *source, char *buf)  
{
    char *cmd, *s;
    char orig[BUFSIZE];

    /* Make a copy of the original message. */ 
    strscpy (orig, buf, sizeof (orig));   

    cmd = strtok (buf, " ");

    if (!cmd)
	return;
    
    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (s_StatServ, source, "PING %s", s);
	return;
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
    {
	ctcpreply (s_StatServ, source, "VERSION OperStats v%s", version);
	return;
    }

    /* A CTCP we don't recognize.. Ignore it. */
    else if (*cmd == '\1')
	return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP",		H_NONE,		do_help},
	    {"STATS",		H_NONE,		do_stats},
	    {"TLD",		H_NONE,		do_tld},
	    {"TLDMAP",		H_NONE,		do_tldmap},
	    {"TLDSTATS",	H_NONE,		do_tldstats},
	    {"ZONE",		H_NONE,		do_zone},
	    {"MAP",		H_NONE,		do_map},
	    {"SERVERS",		H_NONE,		do_servers},
	    {"SEARCH",		H_IRCOP,	sh_search},
	    {"AUTH",		H_IRCOP,	sh_auth},
	    {"DEAUTH",		H_SRA,		sh_deauth},
	    {"SHUTDOWN",	H_SRA,		sh_shutdown},
	    {"RESTART",		H_SRA,		sh_restart},
#ifdef RAWINJECT
	    {"RAW",		H_SRA,		sh_raw},
	    {"INJECT",		H_SRA,		sh_inject},
#endif
	    {"DEBUG",		H_SRA,		sh_debug},
	    {"UPDATE",		H_SRA,		sh_update},
	    {"REHASH",		H_SRA,		sh_rehash},
	    {"SETTINGS",	H_SRA,		sh_settings},
	    {"SENDPINGS",	H_SRA,		do_sendpings},
	    {"FLUSH",		H_SRA,		do_flush},
	    {"RESET",		H_SRA,		do_reset},
	    {NULL}
	};
            
	if ((command = get_hash (s_StatServ, source, strupper (cmd),
				 hash_table)))
	{
	    /* These are only available if OperServ is OFF. */
	    if ((command->process == sh_auth ||
		 command->process == sh_deauth ||
#ifdef RAWINJECT
		 command->process == sh_raw ||
		 command->process == sh_inject ||
#endif
		 command->process == sh_debug ||
		 command->process == sh_shutdown ||
		 command->process == sh_restart ||
		 command->process == sh_settings ||
		 command->process == sh_update ||
		 command->process == sh_search ||
		 command->process == sh_rehash) &&
		 operserv_on == TRUE)
	    {
		notice (s_StatServ, source, ERR_UNKNOWN_COMMAND,
		    strupper (cmd), haveserv_on == TRUE ? "" : "MSG ",
		    s_StatServ, haveserv_on == TRUE ? "" : securitysetting == 1
		    ? "@" : "", haveserv_on == TRUE ? "" : securitysetting == 1
		    ? me.name : "");

		return;
	    }

	    (*command->process) (source);

	    /* If OperServ is off, we want to snoop several commands. */
	    if ((command->process == sh_deauth ||
#ifdef RAWINJECT
		 command->process == sh_raw ||
		 command->process == sh_inject ||
#endif
		 command->process == sh_debug ||
		 command->process == sh_update ||
		 command->process == sh_rehash ||
		 command->process == sh_search ||
		 command->process == sh_settings) &&
		 operserv_on == FALSE)
	    {
		log ("StatServ: <%s> %s", source, orig);
		snoop (s_StatServ, "<%s> %s", source, orig);
	    }
	    if (command->process == sh_auth)
	    {
		log ("StatServ: <%s> AUTH ...", source);
		snoop (s_StatServ, "<%s> AUTH ...", source);
	    }
	}
    }
}

/* Return a help message. */
static void do_help (const char *source) 
{
    char *cmd = strtok (NULL, " ");
            
    if (!cmd)
	statserv_help_index (source);
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"STATS",		H_NONE,		statserv_help_stats},
	    {"TLD",		H_NONE,		statserv_help_tld},
	    {"TLDMAP",		H_NONE,		statserv_help_tldmap},
	    {"TLDSTATS",	H_NONE,		statserv_help_tldstats},
	    {"ZONE",		H_NONE,		statserv_help_zone},
	    {"MAP",		H_IRCOP,	statserv_help_map},
	    {"SERVERS",		H_IRCOP,	statserv_help_servers},
	    {"SEARCH",		H_IRCOP,	shared_help_search},
	    {"AUTH",		H_IRCOP,	shared_help_auth},
	    {"DEAUTH",		H_SRA,		shared_help_deauth},
	    {"SHUTDOWN",	H_SRA,		shared_help_shutdown},
	    {"RESTART",		H_SRA,		shared_help_restart},
#ifdef RAWINJECT
	    {"RAW",		H_SRA,		shared_help_raw},
	    {"INJECT",		H_SRA,		shared_help_inject},
#endif
	    {"DEBUG",		H_SRA,		shared_help_debug},
	    {"UPDATE",		H_SRA,		shared_help_update},
	    {"REHASH",		H_SRA,		shared_help_rehash},
	    {"SETTINGS",	H_SRA,		shared_help_settings},
	    {"SENDPINGS",	H_SRA,		statserv_help_sendpings},
	    {"FLUSH",		H_SRA,		statserv_help_flush},
	    {"RESET",		H_SRA,		statserv_help_reset},
	    {NULL}
	};

	if ((command = get_help_hash (s_StatServ, source, strupper (cmd),
				      hash_table)))
	{
	    /* Hide disabled or non applicable commands. */
	    if (((command->process == statserv_help_tld) && tldlookup == FALSE) ||
		((command->process == statserv_help_tldmap) && tldlookup == FALSE) ||
		((command->process == statserv_help_tldstats) && tldlookup == FALSE) ||
		((command->process == statserv_help_zone) && zonelookup == FALSE) ||
		((command->process == shared_help_search) && operserv_on == TRUE) ||
		((command->process == shared_help_auth) && operserv_on == TRUE) ||
		((command->process == shared_help_deauth) && operserv_on == TRUE) ||
		((command->process == shared_help_shutdown) && operserv_on == TRUE) ||
		((command->process == shared_help_restart) && operserv_on == TRUE) ||
#ifdef RAWINJECT
		((command->process == shared_help_raw) && operserv_on == TRUE) ||
		((command->process == shared_help_inject) && operserv_on == TRUE) ||
#endif
		((command->process == shared_help_debug) && operserv_on == TRUE) ||
		((command->process == shared_help_update) && operserv_on == TRUE) ||
		((command->process == shared_help_rehash) && operserv_on == TRUE) ||
		((command->process == shared_help_settings) && operserv_on == TRUE))
	    {
		notice (s_StatServ, source, ERR_NO_HELP, strupper (cmd));
		return;
	    }

	    (*command->process) (source);
	}
    }
}

/* Show some basic stats. */
static void do_stats (const char *source)
{
    char *param = strtok (NULL, " ");

    if (param && is_oper (source) && !strchr (param, '+'))
    {
	ServStat *serverstat;

	serverstat = findserv (param);

	if (!serverstat)
	{
	    notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	    return;
	}

	notice (s_StatServ, source, SS_SERVSTATS_HEADER, serverstat->name);
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_SERVSTATS_SINCE, get_time
	    (serverstat->statssince, 4), time_ago (serverstat->statssince));
	if (serverstat->splits)
	    notice (s_StatServ, source, SS_SERVSTATS_LASTSPLIT, get_time
		(serverstat->lastsplit, 4), time_ago (serverstat->lastsplit));
	notice (s_StatServ, source, SS_SERVSTATS_SPLITS, serverstat->splits,
	    serverstat->is_split ? " (Split)" : "");
	if (serverstat->lastquit)
	    notice (s_StatServ, source, SS_SERVSTATS_LASTQUIT,
		serverstat->lastquit);
	if (serverstat->clients)
	    notice (s_StatServ, source, SS_SERVSTATS_USERS, serverstat->clients);
	if (serverstat->maxclients)
	    notice (s_StatServ, source, SS_SERVSTATS_MAXUSERS, serverstat->maxclients);
	if (serverstat->maxtime)
	    notice (s_StatServ, source, SS_SERVSTATS_MAXTIME, get_time
		(serverstat->maxtime, 1));
	if (serverstat->lag.tv_sec == 999)
	    notice (s_StatServ, source, SS_SERVSTATS_LAG, "999");
	else
	    notice (s_StatServ, source, SS_SERVSTATS_LAG,
		tv2ms (&(serverstat->lag)));
	if (timerisset (&(serverstat->peaklag)))
	    notice (s_StatServ, source, SS_SERVSTATS_PEAKLAG,
		tv2ms (&(serverstat->peaklag)));
	if (serverstat->is_split)
	{
	    if (serverstat->penalize)
		notice (s_StatServ, source, SS_SERVSTATS_DOWNTIME, time_ago
		    (time (NULL) - (serverstat->splittime + (time (NULL) -
		    serverstat->lastsplit))));
	    else
		notice (s_StatServ, source, SS_SERVSTATS_DOWNTIME, time_ago
		    (time (NULL) - serverstat->splittime));
	    notice (s_StatServ, source, SS_SERVSTATS_SPLITFOR, time_ago
		(serverstat->lastsplit));
	}
	else
	    notice (s_StatServ, source, SS_SERVSTATS_DOWNTIME, time_ago
		(time (NULL) - serverstat->splittime));
	if (serverstat->lastdowntime)
	    notice (s_StatServ, source, SS_SERVSTATS_LASTDOWN, duration
		(serverstat->lastdowntime, 2));
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_FOOTER);
    }
    else
    {
	long uptime;

	uptime = (time (NULL) - me.since);

	/* Make sure they're not giving us bogus tokens */
	if (param && (!strchr (param, 'C') && !strchr (param, 'c') &&
		      !strchr (param, 'M') && !strchr (param, 'm') &&
		      !strchr (param, 'N') && !strchr (param, 'n') &&
		      !strchr (param, 'U') && !strchr (param, 'u') &&
		      !strchr (param, 'D') && !strchr (param, 'd')))
	{
	    notice (s_StatServ, source, SS_BAD_TOKEN, param);
	    return;
	}

	notice (s_StatServ, source, SS_STATS_HEADER);
	notice (s_StatServ, source, " ");

	if (!param || (param && strchr (param, 'C')) ||
	    (param && strchr (param, 'c')))
	    notice (s_StatServ, source, SS_STATS_CURUSER, userstat, countops ());

	if (!param || (param && strchr (param, 'M')) ||
	    (param && strchr (param, 'm')))
	    notice (s_StatServ, source, SS_STATS_MAXUSER, maxusercnt);

	if (!param || (param && strchr (param, 'C')) ||
	    (param && strchr (param, 'c')))
	{
	    if (is_oper (source))
		notice (s_StatServ, source, SS_STATS_CURCHANOPER,
		    countchans (0), countchans (1));
	    else
		notice (s_StatServ, source, SS_STATS_CURCHAN, countchans (0));
	}

	if (!param || (param && strchr (param, 'M')) ||
	    (param && strchr (param, 'm')))
	    notice (s_StatServ, source, SS_STATS_MAXCHAN, maxchancnt);

	if (!param || (param && strchr (param, 'C')) ||
	    (param && strchr (param, 'c')))
	    notice (s_StatServ, source, SS_STATS_CURSERV, servstat);

	if (!param || (param && strchr (param, 'M')) ||
	    (param && strchr (param, 'm')))
	{
	    notice (s_StatServ, source, SS_STATS_MAXSERV, maxservcnt);
	    notice (s_StatServ, source, SS_STATS_MAXHITS, maxhitcnt);
	}

	if (!param || (param && strchr (param, 'N')) ||
	    (param && strchr (param, 'n')))
	{
	    /* If anyone can make this smaller, I'm listening ;) */
	    if (speedsample > 1000000000)
		notice (s_StatServ, source, SS_STATS_CURSPEEDGB, (float)
		    speedsample / (float) 1000000000, (float) speedstat /
		    (float) (speedstat > 1000000000 ? 1000000000 : speedstat >
		    1000000 ? 1000000 : speedstat > 1000 ? 1000 : 1),
		    speedstat > 1000000000 ? "GB" : speedstat > 1000000 ? "MBs"
		    : speedstat > 1000 ? "KB" : "bytes");
	    if (speedsample > 1000000)
		notice (s_StatServ, source, SS_STATS_CURSPEEDMB, (float)
		    speedsample / (float) 1000000, (float) speedstat /
		    (float) (speedstat > 1000000000 ? 1000000000 : speedstat >
		    1000000 ? 1000000 : speedstat > 1000 ? 1000 : 1),
		    speedstat > 1000000000 ? "GB" : speedstat > 1000000 ? "MB"
		    : speedstat > 1000 ? "KB" : "bytes");
	    else if (speedsample > 1000)
		notice (s_StatServ, source, SS_STATS_CURSPEEDKB, (float)
		    speedsample / (float) 1000, (float) speedstat /
		    (float) (speedstat > 1000000000 ? 1000000000 : speedstat >
		    1000000 ? 1000000 : speedstat > 1000 ? 1000 : 1),
		    speedstat > 1000000000 ? "GB" : speedstat > 1000000 ? "MB"
		    : speedstat > 1000 ? "KB" : "bytes");
	    else
		notice (s_StatServ, source, SS_STATS_CURSPEEDBYTES, speedsample,
		    (float) speedstat / (float) (speedstat > 1000000000 ?
		    1000000000 : speedstat > 1000000 ? 1000000 : speedstat >
		    1000 ? 1000 : 1), speedstat > 1000000000 ? "GB" : speedstat
		    > 1000000 ? "MBs" : speedstat > 1000 ? "KB" : "bytes");
	}

	if (!param || (param && strchr (param, 'U')) ||
	    (param && strchr (param, 'u')))
	{
	    notice (s_StatServ, source, SS_STATS_UPTIME, uptime/86400,
		uptime/86400 == 1 ? "" : "s", (uptime/3600) % 24,
		(uptime/60) %60, uptime % 60);

	    /* If the current uptime is higher than the maxuptime, update! */
	    if (uptime > maxuptime)
		maxuptime = uptime;

	    notice (s_StatServ, source, SS_STATS_MAXUPTIME, maxuptime/86400,
		maxuptime/86400 == 1 ? "" : "s", (maxuptime/3600) % 24,
		(maxuptime/60) % 60, maxuptime % 60);
	}

	if (!param || (param && strchr (param, 'C')) ||
	    (param && strchr (param, 'c')))
	    notice (s_StatServ, source, SS_STATS_CURTIME, get_time (time (NULL),
		1));

	if (!param || (param && strchr (param, 'M')) ||
	    (param && strchr (param, 'm')))
	{
	    notice (s_StatServ, source, SS_STATS_MAXUSERTIME, get_time
		(maxusertime, 1));
	    notice (s_StatServ, source, SS_STATS_MAXCHANTIME, get_time
		(maxchantime, 1));
	    notice (s_StatServ, source, SS_STATS_MAXSERVTIME, get_time
		(maxservtime, 1));
	}

	if (!param || (param && strchr (param, 'D')) ||
	    (param && strchr (param, 'd')))
	{
	    /* We only want this blank line here if theres something else
	       above.
	     */
	    if (!param || (param && (strchr (param, 'C') || strchr (param, 'c')
			    ||  strchr (param, 'M') || strchr (param, 'm')
			    ||  strchr (param, 'N') || strchr (param, 'n')
			    ||  strchr (param, 'U') || strchr (param, 'u'))))
		notice (s_StatServ, source, " ");
	    notice (s_StatServ, source, SS_STATS_DAILY, get_time (time (NULL),
		0));
	    notice (s_StatServ, source, " ");
	    notice (s_StatServ, source, SS_STATS_DAILY_HITS, dailyhitcnt);
	    notice (s_StatServ, source, SS_STATS_DAILY_USERS, dailyuserstat,
		get_time (dailyusertime, 2));
	    notice (s_StatServ, source, SS_STATS_DAILY_CHANS, dailychanstat,
		get_time (dailychantime, 2));
	    notice (s_StatServ, source, SS_STATS_DAILY_SERVS, dailyservstat,
		get_time (dailyservtime, 2));
	}

	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_FOOTER);
    }
}

/* Show a listing of servers and their version replies. */
static void do_servers (const char *source)
{
    int i;
    ServStat *server;

    notice (s_StatServ, source, SS_SERVERS_HEADER);
    notice (s_StatServ, source, " ");

    if (ignoreulines_on == FALSE)
    {
	int tmp = servcount ();

	notice (s_StatServ, source, "%s (operstats-%s. %s%s%s%s%s) (%d Max Users) (0 Splits)",
	    me.name, version,
	    (operserv_on == TRUE) ? "O" : "",
	    (statserv_on == TRUE) ? "S" : "",
	    (globalnoticer_on == TRUE) ? "g" : "",
	    (spamserv_on == TRUE) ? "s" : "",
	    debuglevel ? "d" : "",
	    tmp + 1);
    }

    for (i = 0; i < 1024; i++)
    {
	for (server = statlist[i]; server; server = server->next)
	{
	    if (finduline (server->name) && ignoreulines_on == TRUE)
		continue;

	    notice (s_StatServ, source, "%s (%s %s) (%d Max Users) (%d Split%s)%s%s%s",
		server->name, server->version ? server->version : "Unknown",
		server->flags ? server->flags : "XX", server->maxclients,
		server->splits, server->splits == 1 ? "" : "s",
		server->is_split ? " (" : "",
		server->is_split ? "Split" : "",
		server->is_split ? ")" : "");
	}
    }

    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, SS_SERVERS_FOOTER);
}

/* Display the country/group associated with a given TLD. */
static void do_tld (const char *source)
{
    TLD *tld;
    char *param = strtok (NULL, "");

    if (tldlookup == FALSE)
    {
	notice (s_StatServ, source, ERR_UNKNOWN_COMMAND, "TLD",
	    haveserv_on == TRUE ? "" : "MSG ", s_StatServ, haveserv_on == TRUE ?
	    "" : securitysetting == 1 ? "@" : "", haveserv_on == TRUE ? "" :
	    securitysetting == 1 ? me.name : "");

	return;
    }

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "TLD TopLevelDomain|Country");
	errmoreinfo (s_StatServ, source, "TLD");
	return;
    }

    tld = findtld (param);

    if (!tld)
    {
	notice (s_StatServ, source, SS_TLD_NOT_FOUND, param);
	return;
    }
    else
	notice (s_StatServ, source, SS_TLD_IS, tld->name, tld->desc);
}

/* Show either the user count/percentage for the given TLD, or
   show all active TLDs.
 */
static void do_tldmap (const char *source)
{
    TLD *tld;
    char *param = strtok (NULL, " ");

    if (tldlookup == FALSE)
    {
	notice (s_StatServ, source, ERR_UNKNOWN_COMMAND, "TLDMAP",
	    haveserv_on == TRUE ? "" : "MSG ", s_StatServ, haveserv_on == TRUE ? "" :
	    securitysetting == 1 ? "@" : "", haveserv_on == TRUE ? "" :
	    securitysetting == 1 ? me.name : "");

	return;
    }

    if (!param)
    {
	notice (s_StatServ, source, SS_TLDMAP_START);
	notice (s_StatServ, source, " ");

	for (tld = tldlist; tld; tld = tld->next)
	    if (tld->cnt)
		notice (s_StatServ, source, SS_TLDMAP, tld->desc,
		    " (", tld->name, ")", tld->cnt, tld->cnt == 1 ? "" : "s",
		    (float) (tld->cnt * 100) / usercnt);

	if (tldnores)
	    notice (s_StatServ, source, SS_TLDMAP, "Non Resolving", "", "",
		"", tldnores, tldnores == 1 ? "" : "s",
		(float) (tldnores * 100) / usercnt);

	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_TLDMAP_END);
    }
    else
    {
	tld = findtld (param);

	if (!tld)
	{
	    notice (s_StatServ, source, SS_TLD_NOT_FOUND, param);
	    return;
	}
	else
	    notice (s_StatServ, source, SS_TLDMAP, tld->desc, " (", tld->name,
		")", tld->cnt, tld->cnt == 1 ? "" : "s",
		(float) (tld->cnt * 100) / usercnt);
    }
}

/* Show the per-tld daily hit counts. */
static void do_tldstats (const char *source)
{
    TLD *tld;
    int i, j = 0;
    char *param = strtok (NULL, " ");

    if (tldlookup == FALSE)
    {
	notice (s_StatServ, source, ERR_UNKNOWN_COMMAND, "TLDSTATS",
	    haveserv_on == TRUE ? "" : "MSG ", s_StatServ, haveserv_on == TRUE ? "" :
	    securitysetting == 1 ? "@" : "", haveserv_on == TRUE ? "" :
	    securitysetting == 1 ? me.name : "");

	return;
    }

    if (!param)
    {
	notice (s_StatServ, source, SS_TLDSTATS_START,
	    get_time (time (NULL), 0));

	for (tld = tldlist; tld; tld = tld->next)
	    if (tld->hits)
	    {
		i = (tld->hits * 100) / dailyhitcnt;
		j++;
		notice (s_StatServ, source, SS_TLDSTATS, tld->desc, tld->name,
		    i, tld->hits, tld->hits == 1 ? "" : "s");
	    }

	if (tldnoreshits)
	{
	    i = (tldnoreshits * 100) / dailyhitcnt;
	    j++;
	    notice (s_StatServ, source, SS_TLDSTATS, "Non Resolving",
		"Unknown", i, tldnoreshits, tldnoreshits == 1 ? "" : "s");
	}

	notice (s_StatServ, source, SS_TLDSTATS_END, dailyhitcnt, j);
    }
    else
    {
	tld = findtld (param);

	if (!tld)
	{
	    notice (s_StatServ, source, SS_TLD_NOT_FOUND, param);
	    return;
	}
	else
	{
	    i = (tld->hits * 100) / dailyhitcnt;
	    notice (s_StatServ, source, SS_TLDSTATS, tld->desc, tld->name,
		i, tld->hits, tld->hits == 1 ? "" : "s");
	}
    }
}

/* Display the current time and/or GMT offset of the given timezone. */
static void do_zone (const char *source)
{
    Zone *zone;
    char *param = strtok (NULL, " ");
    int found = 0;

    if (zonelookup == FALSE)
    {
	notice (s_StatServ, source, ERR_UNKNOWN_COMMAND, "ZONE",
	    haveserv_on == TRUE ? "" : "MSG ", s_StatServ, haveserv_on == TRUE ? "" :
	    securitysetting == 1 ? "@" : "", haveserv_on == TRUE ? "" :
	    securitysetting == 1 ? me.name : "");

	return;
    }

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "ZONE TimeZone");
	errmoreinfo (s_StatServ, source, "ZONE");
	return;
    }

    /* Search the TimeZone list, returning matches as they're found.
       Note that this doesn't call an external function: Since several
       TimeZones may share the same abbreviation, we'll search the list
       and return all matches.
     */
    for (zone = zonelist; zone; zone = zone->next)
    {
        if (!stricmp (param, zone->name))
	{
	    notice (s_StatServ, source, SS_ZONE_IS, zone_time (zone->offset),
		zone->desc, zone->name, zone->noffset);
	    found = 1;
	}
    }

    /* If we didn't find anything, let them know. */
    if (!found)
    {
	notice (s_StatServ, source, SS_ZONE_NOT_FOUND, strupper (param));
	return;
    }
}

/* Reset a given server's stats */
static void do_reset (const char *source)
{
    ServStat *server;
    char *param = strtok (NULL, " ");

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "RESET Server|ALL");
	errmoreinfo (s_StatServ, source, "RESET");
	return;
    }

    /* Reset a specific server */
    if (stricmp (param, "ALL"))
    {
	server = findserv (param);

	if (!server)
	{
	    notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	    return;
	}
	else
	{
	    time_t now = time (NULL);

	    server->statssince = now;
	    server->lastsplit = now;
	    server->maxtime = now;
	    server->maxclients = server->clients;
	    server->lastdowntime = 0;
	    server->peaklag.tv_sec = 0;
	    server->peaklag.tv_usec = 0;
	    server->splittime = 0;

	    if (server->is_split)
		server->splits = 1;
	    else
		server->splits = 0;

	    if (server->lastquit)
		free (server->lastquit);

	    server->lastquit = "N/A";

	    notice (s_StatServ, source, SS_SERVER_RESET, server->name);
	}
    }

    /* Reset ALL servers */
    else
    {
	int i;

	for (i = 0; i < 1024; i++)
	{
	    for (server = statlist[i]; server; server = server->next)
	    {
		time_t now = time (NULL);

		server->statssince = now;
		server->lastsplit = now;
		server->maxtime = now;
		server->maxclients = server->clients;
		server->lastdowntime = 0;
		server->peaklag.tv_sec = 0;
		server->peaklag.tv_usec = 0;
		server->splittime = 0;

		if (server->is_split)
		    server->splits = 1;
		else
		    server->splits = 0;

		if (server->lastquit)
		    free (server->lastquit);

		server->lastquit = "N/A";
	    }
	}

	notice (s_StatServ, source, SS_SERVER_RESET, "all servers");
    }
}

/* FLUSH: Remove a server from the stats list (ie delinked) */
static void do_flush (const char *source)
{
    ServStat *server;
    char *param = strtok (NULL, " ");

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "FLUSH Server");
	errmoreinfo (s_StatServ, source, "FLUSH");
	return;
    }

    server = findserv (param);

    if (!server)
    {
	notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	return;
    }
    else
    {
	int i;

	/* Only flush servers that aren't online. */
	for (i = 0; i < servcnt; i++)
	{
	    if (!stricmp (servlist[i].server, param))
	    {
		notice (s_StatServ, source, SS_CANT_FLUSH);
		return;
	    }
	}

	notice (s_StatServ, source, SS_SERVER_FLUSHED, server->name);

	if (server->version)
	    free (server->version);

	if (server->flags)
	    free (server->flags);

	if (server->lastquit)
	    free (server->lastquit);

	if (server->prev)
	    server->prev->next = server->next;
	else
	    statlist[SHASH(server->name)] = server->next;
	if (server->next)
	    server->next->prev = server->prev;

	if (server->name)
	    free (server->name);

	free (server);
    }
}

/* Show a map of the network. */
static void do_map (const char *source)
{
    int i;
    char l[64];

    /* Find out which server we are connected to */
    for (i = 0; i < servcnt; i++)
	if (servlist[i].hops == 1)
	    snprintf (l, sizeof (l), "%s", servlist[i].server);

    notice (s_StatServ, source, SS_MAP_HEADER);
    notice (s_StatServ, source, " ");
    showleafs (source, l, "", "");
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, SS_MAP_FOOTER);
}

/* Associated MAP function */
int showleafs (const char *source, char *server, char *prevpre,
	       char *arrowchar)
{
    User *user;
    ServStat *serverstat;
    int i = 0, j = 0, k = 0, x, num = 0, cnt, opcnt;
    char buf[BUFSIZE];
    char pre[BUFSIZE];
    char *nextpre;
    static int depth = 0;
 
    nextpre = malloc (BUFSIZE);
   
    snprintf (pre, sizeof (pre) + 1, prevpre);

    *buf = 0;

    for (i = 0; i < servcnt; i++)
    {
	cnt = 0;
	opcnt = 0;

	/* Hide U:Lined servers. */
	if (finduline (servlist[i].server) && ignoreulines_on == TRUE)
	    if (!is_oper (source))
		continue;

	/* Also hide servers with ulined uplinks (aka, jupes) */
	if (finduline (servlist[i].uplink) && ignoreulines_on == TRUE)
	    if (!is_oper (source))
		continue;

	serverstat = findserv (servlist[i].server);

	for (x = 0; x < 1024; x++)
	    for (user = userlist[x]; user; user = user->next)
		if (!strcmp (servlist[i].server, user->server))
		{
		    cnt++;

		    if (is_oper (user->nick))
			opcnt++;
		}

	if (servlist[i].hops == 1 && match_wild_nocase (server, servlist[i].server))
	{
	    if (serverstat->lag.tv_sec == 999)
	    {
		notice (s_StatServ, source,
		    "\2%s\2 (999ms) (%d user%s, (%d op%s), %3.2f%%)",
		    servlist[i].server, cnt, cnt == 1 ? "" : "s", opcnt,
		    opcnt == 1 ? "" : "s", ((float) (cnt * 100) / usercnt));

		continue;
	    }

	    if (tv2ms (&(serverstat->lag)) >= (lagthreshold * 1000))
		notice (s_StatServ, source,
		    "\2%s\2 (%dms) (%d user%s (%d op%s), %3.2f%%)",
		    servlist[i].server, tv2ms (&(serverstat->lag)), cnt,
		    cnt == 1 ? "" : "s", opcnt, opcnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	    else
		notice (s_StatServ, source,
		    "%s (%dms) (%d user%s (%d op%s), %3.2f%%)",
		    servlist[i].server, tv2ms (&(serverstat->lag)), cnt,
		    cnt == 1 ? "" : "s", opcnt, opcnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	}
	else if (match_wild_nocase (server, servlist[i].server))
	{
	    strcat (buf, pre);
	    strcat (buf, arrowchar);
	    strcat (buf, "-");

	    if (tv2ms (&(serverstat->lag)) > 10000)
		strcat (buf, "\2");

	    strcat (buf, servlist[i].server);

	    if (serverstat->lag.tv_sec == 999)
	    {
		notice (s_StatServ, source,
		    "%s\2 (999ms) (%d user%s, (%d op%s), %3.2f%%)",
		    buf, cnt, cnt == 1 ? "" : "s", opcnt,
		    opcnt == 1 ? "" : "s", ((float) (cnt * 100) / usercnt));

		continue;
	    }

	    if (tv2ms (&(serverstat->lag)) >= (lagthreshold * 1000))
		notice (s_StatServ, source,
		    "%s\2 (%dms) (%d user%s (%d op%s), %3.2f%%)",
		    buf, tv2ms (&(serverstat->lag)), cnt, cnt == 1 ? "" : "s",
		    opcnt, opcnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	    else
		notice (s_StatServ, source,
		    "%s (%dms) (%d user%s (%d op%s), %3.2f%%)",
		    buf, tv2ms (&(serverstat->lag)), cnt, cnt == 1 ? "" : "s",
		    opcnt, opcnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	}
    }

    j = numleafs (server);
    k = depth;

    for (i = 0; i < servcnt; i++)
    {
	depth = k;

	if (!stricmp (servlist[i].uplink, server))
	{
	    snprintf (nextpre, strlen (pre) + 1, pre);

	    if (depth++ > 0)
	    {
		if (arrowchar[0] == '`')
		    if (servlist[i].hops == 1)
			strcat (nextpre, "    ");
		    else
			strcat (nextpre, "  ");
		else
		    if (servlist[i].hops == 1)
			strcat (nextpre, "  | ");
		    else
			strcat (nextpre, "| ");
	    }

	    if (j > ++num)
		if (servlist[i].hops == 1)
		    showleafs (source, servlist[i].server, nextpre, "  |");
		else
		    showleafs (source, servlist[i].server, nextpre, "|");
	    else
		if (servlist[i].hops == 1)
		    showleafs (source, servlist[i].server, nextpre, "  `");
		else
		    showleafs (source, servlist[i].server, nextpre, "`");
	}
    }
    free (nextpre);
    depth = 0;
    return 1;
}

/* Associated MAP function */
int numleafs (char *server)
{
    int i, num = 0;
         
    for (i = 0; i < servcnt; i++)
    {
	if (!stricmp (servlist[i].uplink, server))
	num++;
    }        
    return (num);
}

/* Load the StatServ DB from disk */
void load_ss_dbase (void)
{
    FILE *f = fopen (STATSERV_DB, "r");
    char *item, dBuf[2048], backup[2048];
    int convert = 0;

    log (RPL_LOADING, STATSERV_DB);

    if (!f)
    {
	log ("Warning: No %s found.", STATSERV_DB);

	/* Check for a db.save */
	strcpy (backup, STATSERV_DB);
	strcat (backup, ".save");
	if (rename (backup, STATSERV_DB) > 0)
	{
	    log ("Recovering %s.save", STATSERV_DB);
	    f = fopen (STATSERV_DB, "r");
	}
	else
	    return;
    }
    
    while (fgets (dBuf, 2047, f))
    {
	item = strtok (dBuf, " ");

	/* Ignore certain things.. */
	if (*item == '#' || *item == '\n' || *item == '\t' || *item == ' ' ||
	    *item == '\0' || *item == '\r')
	    continue;

	if (!strcmp (item, "SV"))
	{
	    int dbver = atoi (strtok (NULL, " "));

	    if (dbver < ss.version)
	    {
		convert = 1;
		log ("Importing StatServ v%d database to v%d format", dbver, ss.version);
	    }
	}

	if (!strcmp (item, "ST"))
	{
	    maxusercnt	= atoi (strtok (NULL, " "));
	    maxusertime	= atol (strtok (NULL, " "));
	    maxchancnt	= atoi (strtok (NULL, " "));
	    maxchantime	= atol (strtok (NULL, " "));
	    maxservcnt	= atoi (strtok (NULL, " "));
	    maxservtime	= atol (strtok (NULL, " "));
	    maxhitcnt	= atoi (strtok (NULL, " "));

	    if (ss.version > 3 && !convert)
		maxuptime = atoi (strtok (NULL, " "));

	    speedstat	= atoi (strtok (NULL, "\n"));

	    continue;
	}

	if (!strcmp (item, "SD"))
	{
	    dailyuserstat = atoi (strtok (NULL, " "));
	    dailyusertime = atol (strtok (NULL, " "));
	    dailychanstat = atoi (strtok (NULL, " "));
	    dailychantime = atol (strtok (NULL, " "));
	    dailyservstat = atoi (strtok (NULL, " "));
	    dailyservtime = atol (strtok (NULL, "\n"));

	    continue;
	}

	if (!strcmp (item, "SS"))
	{
	    ServStat *serverstat, **list;

	    serverstat = scalloc (sizeof (ServStat), 1);

	    serverstat->name = sstrdup (strtok (NULL, " "));
	    serverstat->statssince = atol (strtok (NULL, " "));
	    serverstat->lastsplit = atol (strtok (NULL, " ")); 
	    serverstat->is_split = atoi (strtok (NULL, " "));
	    serverstat->splittime = atol (strtok (NULL, " "));
	    serverstat->splits = atoi (strtok (NULL, " "));
	    serverstat->lastdowntime = atoi (strtok (NULL, " "));

	    if (ss.version > 1 && !convert)
	    {
		serverstat->is_uline = atoi (strtok (NULL, " "));
		serverstat->peaklag.tv_sec = atoi (strtok (NULL, " "));
		serverstat->peaklag.tv_usec = atoi (strtok (NULL, " "));
	    }

	    if (ss.version > 2 && !convert)
	    {
		serverstat->maxtime = atol (strtok (NULL, " "));
		serverstat->maxclients = atoi (strtok (NULL, " "));
	    }

	    serverstat->lastquit = sstrdup (strtok (NULL, "\n"));

	    list = &statlist[SHASH(serverstat->name)];
	    serverstat->next = *list;
	    if (*list)
		(*list)->prev = serverstat;
	    *list = serverstat;

	    continue;
	}
    }

    if (convert)
	log ("Conversion to StatServ v%d database was successful.", ss.version);

    fclose (f);
}

/* Write the stats db to disk. */
void save_ss_dbase ()
{
    FILE *f;
    ServStat *server;
    int i;
    char backup[2048];

    if (logupdates_on == TRUE)
	log (RPL_SAVING, STATSERV_DB);

    strcpy (backup, STATSERV_DB);  
    strcat (backup, ".save");
    remove (backup);
    if (rename (STATSERV_DB, backup) < 0)
	log ("Can't back up %s!", STATSERV_DB);
    f = fopen (STATSERV_DB, "w");
    
    if (!f)
    {
	log ("Can't write to %s!", STATSERV_DB);
	if (rename (backup, STATSERV_DB) < 0)
	    fatal ("Can't restore backup of %s!", STATSERV_DB);
	return;
    }

    /* Write the version to the DB */
    fprintf (f, "SV %d\n", ss.version);

    /* Update maxuptime before we save it. Rather than check every second if it's
       a new record, we only check when someone request ss stats, and here. This
       will cause us to lose up to synctime of uptime in the invent of an improper
       shutdown, but this is acceptable to me.
     */
    if ((time (NULL) - me.since) > maxuptime)
	maxuptime = (time (NULL) - me.since);

    /* Various Statistics */
    fprintf (f, "ST %d %lu %d %lu %d %lu %d %li %d\n", maxusercnt, maxusertime,
	maxchancnt, maxchantime, maxservcnt, maxservtime, maxhitcnt, maxuptime,
	speedstat);

    /* Stats for the current day */
    fprintf (f, "SD %d %lu %d %lu %d %lu\n", dailyuserstat, dailyusertime,
	dailychanstat, dailychantime, dailyservstat, dailyservtime);

    /* Per-Server stats */
    for (i = 0; i < 1024; i++)
    {
	for (server = statlist[i]; server; server = server->next)
	{
	    fprintf (f, "SS %s %lu %lu %d %d %d %d %d %ld %ld %lu %d %s\n", server->name,
		server->statssince, server->lastsplit, server->is_split,
		server->splittime, server->splits, server->lastdowntime,
		server->is_uline, server->peaklag.tv_sec, server->peaklag.tv_usec,
		server->maxtime, server->maxclients, server->lastquit ?
		server->lastquit : "N/A");
	}
    }

    fclose (f);
    remove (backup);
}

/* Add a new TLD into the tld list. */
static TLD * new_tld (const char *name, const char *desc)
{
    TLD *tld;

    tld = scalloc (sizeof (TLD), 1);

    tld->name = sstrdup (name);
    tld->desc = sstrdup (desc);

    tld->next = tldlist;

    if (tld->next)
	tld->next->prev = tld;

    tldlist = tld;

    return tld;
}

/* Load the TLD listings */
void load_tlds (void)
{
    FILE *f = fopen (TLDFILE, "r");
    char *s, dBuf[BUFSIZE], *name, *desc;

    log (RPL_LOADING, TLDFILE);

    if (!f)
    {
	tldlookup = FALSE;
	dumptlds_on = FALSE;
	return;
    }

    while (fgets (dBuf, 2047, f))
    {
	s = strtok (dBuf, " ");
	strip (s);

	if (s)
	    name = s;
	else
	    name = "";

	s = strtok (NULL, "");
	strip (s);

	if (s)
	    desc = s;
	else
	    desc = "";

	if (!desc)
	    log ("Found TLD %s but no description!", name);
	else
	    new_tld (name, desc);
    }
    fclose (f);
    tldlookup = TRUE;
}

/* Add a new TimeZone into the timezones list. */
static Zone *new_zone (const char *name,  const char *noffset,
		       const char *offset, const char *desc)
{
    Zone *zone;

    zone = scalloc (sizeof (Zone), 1);

    zone->name = sstrdup (name);
    zone->noffset = sstrdup (noffset);
    zone->offset = atoi (offset);
    zone->desc = sstrdup (desc);

    zone->next = zonelist;

    if (zone->next)
	zone->next->prev = zone;

    zonelist = zone;

    return zone;
}

/* Load the TimeZones into memory. */
void load_zones (void)
{
    FILE *f = fopen (ZONEFILE, "r");
    char *s, dBuf[BUFSIZE], *name, *noffset, *offset, *desc;

    log (RPL_LOADING, ZONEFILE);

    if (!f)
    {
	zonelookup = FALSE;
	return;
    }

    while (fgets (dBuf, 2047, f))
    {
	s = strtok (dBuf, " ");
	strip (s);

	if (s)
	    name = s;
	else
	    name = "";

	s = strtok (NULL, " ");
	strip (s);

	if (s)
	    noffset = s;
	else
	    noffset = "";

	s = strtok (NULL, " ");
	strip (s);

	if (s)
	    offset = s;
	else
	    offset = "";

	s = strtok (NULL, "");
	strip (s);

	if (s)
	    desc = s;
	else
	    desc = "";

	if (!desc)
	    log ("Invalid TimeZone %s: Missing data!", name);
	else
	    new_zone (name, noffset, offset, desc);
    }

    fclose (f);
    zonelookup = TRUE;
}

/* Ping all servers. */
void do_sendpings (const char *source)
{
    int i;

    for (i = 0; i < servcnt; i++)
    {
	send_cmd (me.name, "%s %s %s", me.token ? "8" : "PING",
	    me.name, servlist[i].server);
	gettimeofday (&(servlist[i].lastping), NULL);
    }

    notice (s_StatServ, source, SS_PINGED_ALL);
}

/* Reset the daily statistics to current values. */
void reset_daily_stats (void)
{
    TLD *tld;

    dailyuserstat = userstat;
    dailyusertime = time (NULL);
    dailychanstat = chanstat;
    dailychantime = time (NULL);
    dailyservstat = servstat;
    dailyservtime = time (NULL);
    dailyhitcnt = 0;
    tldnoreshits = 0;

    for (tld = tldlist; tld; tld = tld->next)
	if (tld->hits)
	    tld->hits = 0;
}

/* Shared functions: These call shared functions with the appropriate
   information.
 */
static void sh_search (const char *source)
{
    do_search (s_StatServ, source);
}

static void sh_auth (const char *source)
{
    do_auth (s_StatServ, source);
}

static void sh_deauth (const char *source)
{
    do_deauth (s_StatServ, source);
}

#ifdef RAWINJECT
static void sh_raw (const char *source)
{
    do_raw (s_StatServ, source);
}

static void sh_inject (const char *source)
{
    do_inject (s_StatServ, source);
}
#endif

static void sh_debug (const char *source)
{
    do_debug (s_StatServ, source);
}           

static void sh_shutdown (const char *source)
{
    do_shutdown (s_StatServ, source);
}

static void sh_restart (const char *source)
{
    do_restart (s_StatServ, source);
}

static void sh_settings (const char *source)
{
    do_settings (s_StatServ, source);
}

static void sh_update (const char *source)
{
    do_update (s_StatServ, source);
}

static void sh_rehash (const char *source)
{
    do_rehash (s_StatServ, source);
}
